# -*- coding: utf-8 -*-

import arcpy, os
version_major = 1
version_minor = 0
version_bugfix = 0
version_description = "TEDR {}.{}.{}".format(version_major, version_minor, version_bugfix)
version_release = 100
version_table_name = "tedr_version"

xml_workspace_name = 'tedr_1_1_XMLWorkspace.xml'
current_dir = os.path.dirname(__file__)


info_logger = None
err_logger = None
is_arcmap = False
#formatas, kuriuos DB saugoma data
et_date_format = "DATABASE_TIME"


d = arcpy.GetInstallInfo()
if d["ProductName"] == "Desktop":
    is_arcmap = True
    reload(sys)
    sys.setdefaultencoding("utf-8")
else:
    is_arcmap = False

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Toolbox"
        self.alias = "toolbox"

        # List of tool classes associated with this toolbox
        self.tools = [CreateTedr,appendtoTEDR]


class CreateTedr(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""

        if is_arcmap: self.label = "1. Sukurti TEDR struktura"
        else: self.label = "1. Sukurti TEDR struktūrą"
        self.description = "Duomenų bazėje sukuria TEDR reikalingas struktūras"
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        params = None

        param0 = arcpy.Parameter(
            displayName="TEDR duomenų saugykla (tedr.sde)",
            name="tedr_data",
            datatype="DEWorkspace",
            parameterType="Required",
            direction="Input")
        param0.filter.list = ["Remote Database"]
        if is_arcmap: param0.displayName = "TEDR duomenu saugykla (tedr.sde)"

        param1 = arcpy.Parameter(
            displayName="Geoduomenų bazės autorizacijos failas",
            name="authorization_file",
            datatype="DEFile",      
            parameterType="Required",  
            direction="Input")        
        if is_arcmap: param1.displayName = "Geoduomenu bazes autorizacijos failas"

        param2 = arcpy.Parameter(
            displayName="Rezultatas",
            name="result",
            datatype="Double",
            parameterType="Derived",
            direction="Output")

        params = [param0, param1, param2]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""

        result = initTEDR(
            parameters[0].valueAsText,
            parameters[1].valueAsText            
        )
        
        if result is True:
            arcpy.SetParameter(2, 1)
        else:
            arcpy.SetParameter(2, 0)


class appendtoTEDR(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        if is_arcmap: self.label = "2. SEDR2021 FGDB i TEDR DB"
        else: self.label = "2. SEDR2021 FGDB į TEDR DB"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""


        param0 = arcpy.Parameter(
            displayName="Keliamas duomenų rinkinys Esri FGDB formatu (SEDR2021)",
            name="SEDRFGDB2021",
            datatype="DEWorkspace",
            parameterType="Required",
            direction="Input")
        if is_arcmap:
            param0.displayName = param0.displayName.encode('cp1257')

        param1 = arcpy.Parameter(
                displayName="TEDR įmonės DB (SEDR2021 struktūra)",
                name="TEDRDB",
                datatype="DEWorkspace",
                parameterType="Required",
                direction="Input")
        if is_arcmap:
                param1.displayName = "TEDR imones DB (SEDR2021 struktura)"

        params = [param0,param1]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
       
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        appendMatched(
            parameters[0].valueAsText,
            parameters[1].valueAsText
        )







# TEDR sukūrimo funkcijos
def initTEDR(tedr_sde, authorization_file):

    if not arcpy.Exists(tedr_sde):
        # arcpy.AddError("Could not locate TEDR sde file")
        arcpy.AddError("Nenustatytas prisijungimo TEDR sde failas")

    if not arcpy.Exists(authorization_file):
        # arcpy.AddError("Could not locate authorization file")
        arcpy.AddError("Nenustatytas autorizacijos failas")

    new_version_info = (version_description, version_release)
    # arcpy.AddMessage("Database will be upgraded to version: {}".format(new_version_info))
    arcpy.AddMessage("DB bus atnaujinta i versija: {}".format(new_version_info))

    detected_version_info = check_version(tedr_sde)
    # arcpy.AddMessage("Detected version: {}".format(detected_version_info))
    arcpy.AddMessage("Nustatyta versija: {}".format(detected_version_info))

    if detected_version_info[0]:
        # arcpy.AddMessage("Upgrade for this version is not available")       
        arcpy.AddMessage("Atnaujinimas turimai versijai nenustatytas")  
    else:
        create_initial_db_structure(tedr_sde, authorization_file)    
    
    return True

def create_initial_db_structure(tedr_sde, authorization_file):
    # enable enterprise geodatabase
    # arcpy.AddMessage("Enabling enterprise geodatabase")
    arcpy.AddMessage("Igalinamas enterprise geodatabase funcionalumas")
    arcpy.EnableEnterpriseGeodatabase_management(input_database=tedr_sde, authorization_file=authorization_file)
    
    # import xml workspace    
    xml = get_resource_file(xml_workspace_name)
    # arcpy.AddMessage("Importing workspace xml document {}".format(xml))
    arcpy.AddMessage("Importuojamas strukturos XML dokumentas: {}".format(xml))
    arcpy.ImportXMLWorkspaceDocument_management(target_geodatabase=tedr_sde, in_file=xml, import_type="SCHEMA_ONLY")

    # enable editor tracking
    enable_editor_tracking(tedr_sde, 'Nauj_n', 'Nauj_d', 'Red_n', 'Red_d')
    # enable_editor_tracking(	sde_connection_file, 	created_user_field, 	created_date_field, 	last_edited_user_field, 	last_edited_date_field)
    # register as versioned
    # arcpy.AddMessage("Registering as versioned")
    arcpy.AddMessage("Registruojama versijuota DB")
    register_as_versioned(tedr_sde)    

    # add tedr version info
    # arcpy.AddMessage("Adding tedr version info")
    arcpy.AddMessage("Pridedama versijos informacija")
    add_version_info(tedr_sde)

def check_version(sde_connection_file):
    version_table_path = os.path.join(sde_connection_file, version_table_name)
    db_version_release = None
    db_version_description = None
    if arcpy.Exists(version_table_path):
        fields = ['MAJOR', 'MINOR', 'BUGFIX', 'DESCRIPTION', 'RELEASE']
        with arcpy.da.SearchCursor(version_table_path, fields) as cursor:
            for row in cursor:
                db_version_release = row[4]
                db_version_description = row[3]
                break

    return (db_version_description, db_version_release)

def get_resource_file(file_name):
    file_path = os.path.abspath(os.path.join(os.path.dirname( __file__ ), '..','konfiguracija', file_name))
    if not arcpy.Exists(file_path):
        # tool opened from toolbox
        file_path = os.path.abspath(os.path.join(os.path.dirname( __file__ ), '..', '..', 'konfiguracija', file_name))
    return file_path

def add_version_info(sde_connection_file):    
    
    version_table_path = os.path.join(sde_connection_file, version_table_name)
    fields = ['MAJOR', 'MINOR', 'BUGFIX', 'DESCRIPTION', 'RELEASE']

    arcpy.CreateTable_management(sde_connection_file, version_table_name)

    arcpy.AddField_management(version_table_path, "MAJOR", "LONG")
    arcpy.AddField_management(version_table_path, "MINOR", "LONG")
    arcpy.AddField_management(version_table_path, "BUGFIX", "LONG")
    arcpy.AddField_management(version_table_path, "DESCRIPTION", "Text", 96)
    arcpy.AddField_management(version_table_path, "RELEASE", "LONG")

    cursor = arcpy.da.InsertCursor(version_table_path, fields)
    cursor.insertRow((version_major, version_minor, version_bugfix, version_description, version_release))
    del cursor

def register_as_versioned(sde_connection_file):	

	arcpy.env.workspace = sde_connection_file

	# Get all the stand alone tables and feature classes
	dataList = arcpy.ListTables() + arcpy.ListFeatureClasses()
	
	# Add detasets to list
	for dataset in arcpy.ListDatasets("", "Feature"):
		dataList.append(dataset)

	arcpy.AddMessage("Rasti objektu rinkiniai {}".format(dataList))	
	for dataset in dataList:		
		arcpy.AddMessage('Registruojama versijuota ' + dataset)
		arcpy.RegisterAsVersioned_management(dataset, "NO_EDITS_TO_BASE")


def enable_editor_tracking(
	sde_connection_file, 
	created_user_field, 
	created_date_field, 
	last_edited_user_field, 
	last_edited_date_field):
	
	workspace = sde_connection_file
	arcpy.env.workspace = workspace

	# Get all the stand alone tables and feature classes
	standaloneList = arcpy.ListTables() + arcpy.ListFeatureClasses()

	# Execute enable editor tracking on standalone tables
	for dataset in standaloneList:
		arcpy.AddMessage('Ijungiamas redagavimo sekimas ' + dataset)
		arcpy.EnableEditorTracking_management(dataset, created_user_field, created_date_field, last_edited_user_field, last_edited_date_field, "ADD_FIELDS", et_date_format)

	# For feature datasets get all of the featureclasses and add them to the list
	for dataset in arcpy.ListDatasets("", "Feature"):
		arcpy.env.workspace = os.path.join(workspace,dataset)
		for fc in arcpy.ListFeatureClasses():
			arcpy.AddMessage('Ijungiamas redagavimo sekimas ' + fc)
			arcpy.EnableEditorTracking_management(fc, created_user_field, created_date_field, last_edited_user_field, last_edited_date_field, "ADD_FIELDS", et_date_format)




def appendMatched(sourceGDB,targetGDB):
    # randami atitinkantys source ir target FC
    matchedFC = getMatchedFCS(sourceGDB,targetGDB)
    arcpy.AddMessage("Rasta atitinkanciu duomenu rinkiniu:{}".format(len(matchedFC)))
    
    for fc in matchedFC:
        arcpy.AddMessage("Objektu grupe:{}".format(fc[0]))
        fldMappings = getFieldMappings(fc[0],fc[1])
        # reikia išjungti target redagavimo sekimą, nes bus sugadintos datos
        arcpy.AddMessage("Isjungiamas redagavimo sekimas (Editor Tracking): {}".format(fc[1]))
        arcpy.management.DisableEditorTracking(fc[1])
        # perkėlimas
        arcpy.AddMessage("Perkeliami duomenys i: {}".format(fc[1]))
        arcpy.Append_management([fc[0]], fc[1], "NO_TEST", fldMappings)
        # įjungiamas redagavimo sekimas
        arcpy.AddMessage("Ijungiamas redagavimo sekimas (Editor Tracking): {}".format(fc[1]))
        arcpy.management.EnableEditorTracking(fc[1],creator_field='Nauj_n',creation_date_field='Nauj_d',last_editor_field='Red_n',last_edit_date_field='Red_d',record_dates_in=et_date_format)

        

# Įrankio perkelti SEDR2021 į TEDR DB funkcijos
def getMatchedFCS(sourceGDB,targetGDB):
    srcfclist = []
    arcpy.env.workspace = sourceGDB
    datasets = arcpy.ListDatasets(feature_type='feature')
    datasets = [''] + datasets if datasets is not None else []
    for ds in datasets:
        for fc in arcpy.ListFeatureClasses(feature_dataset=ds):
            fc_path = os.path.join(ds, fc)
            srcfclist.append([ds, fc])

    targetfclist = []
    arcpy.env.workspace = targetGDB
    datasets = arcpy.ListDatasets(feature_type='feature')
    datasets = [''] + datasets if datasets is not None else []
    for ds in datasets:
        for fc in arcpy.ListFeatureClasses(feature_dataset=ds):
            fc_path = os.path.join(ds, fc)
            targetfclist.append([ds, fc])

    matched = []
    for srcobj in srcfclist:
        for targetobj in targetfclist:
            if srcobj[0].split(".")[-1] == targetobj[0].split(".")[-1] and srcobj[1].split(".")[-1] == targetobj[1].split(".")[-1]:
                matched.append([os.path.join(sourceGDB,srcobj[0],srcobj[1]),os.path.join(targetGDB,targetobj[0],targetobj[1])])
    return matched


def getFieldMappings(sourceFC, targetFC):
    geomtypeMath = False
    fieldMappings = arcpy.FieldMappings()
    fldMap = arcpy.FieldMap()
    descSRC = arcpy.Describe(sourceFC)
    descTRG = arcpy.Describe(targetFC)
    if descSRC.shapeType == descTRG.shapeType: geomtypeMath = True
    targetFieldNames = [[fld.name,fld.type] for fld in descTRG.fields]
    field_list_map = []
    for srcField in descSRC.fields:
        # if debugmode: arcpy.AddMessage("Fieldmatching. Source Field: {}. Target Fieldlist: {}.".format(srcField.name, targetFieldNames))
        for trgField in targetFieldNames:
            if srcField.name.lower() == trgField[0].lower() and srcField.name.lower() not in ['shape']:
                #čia dar reikia sulyginti tinkamus tipus. 
                # Pavojinga leisti kelti double į small integer, bet �i sąlyga yra. Reikia tikrinti kėlimo metu ar konverutuojasi
                if (srcField.type == trgField[1]) or (srcField.type == 'SmallInteger' and trgField[1] in ['Integer','Double']) or (srcField.type == 'Integer' and trgField[1] in ['Double'])  or (srcField.type == 'Double' and trgField[1] in ['Integer','SmallInteger']):
                    field_list_map.append(srcField.name)
                    fldMap = arcpy.FieldMap()
                    fldMap.addInputField (sourceFC,srcField.name)
                    fldOut = fldMap.outputField
                    fldOut.name = trgField[0]
                    fldOut.name = trgField[1]
                    fieldMappings.addFieldMap(fldMap)
            
    return fieldMappings