# -*- coding: utf-8 -*-

import arcpy
import sys, os, xlrd, json, re, logging, uuid
from codecs import encode, decode
import logging.handlers as handlers
from datetime import datetime
import shutil
import time

info_logger = None
err_logger = None
is_arcmap = False
# if arcpy.GetInstallInfo()["ProductName"] == 'Desktop': is_arcmap = True
gedr_field_types = None
edit_objects = None
last_generated_uid = None
rez_attrib_name = "perkelta"
current_dir = os.path.dirname(__file__)
wsXML_topo = 'tedr_fgdb.xml'
wsXML_betopo = 'tedr_fgdb_betop.xml'
XLSXconfig_name = 'SEDR_TEDR_Taisykles.xlsx'



d = arcpy.GetInstallInfo()
if d["ProductName"] == "Desktop":
    is_arcmap = True
    reload(sys)
    sys.setdefaultencoding("utf-8")
else:
    is_arcmap = False

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "SEDR į SEDR2021 įrankinė"
        self.alias = "SedrToTedrTools"
        self.description = "versija  1.3"        

        # List of tool classes associated with this toolbox
        self.tools = [SedrToTedr]


class SedrToTedr(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "SEDR2018 į SEDR2021"
        if is_arcmap:
            self.label = "SEDR2018 i SEDR2021"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""


        param0 = arcpy.Parameter(
            displayName="SEDR2018 duomenų rinkinys Esri FGDB formatu",
            name="SEDR",
            datatype="DEWorkspace",
            parameterType="Required",
            direction="Input")
        if is_arcmap:
            param0.displayName = "SEDR2018 duomenu rinkinys Esri FGDB formatu"
            param0.displayName = param0.displayName.encode('cp1257')

        param1 = arcpy.Parameter(
            displayName="Gaunamo SEDR2021 rinkinį atitinkančio duomenų failo direktorija",
            name="TEDR_DIR",
            datatype="DEWorkspace",
            parameterType="Required",
            direction="Input")
        if is_arcmap:
            param1.displayName = "Gaunamo SEDR2021 rinkini atitinkancio duomenų failo direktorija"
            param1.displayName = param1.displayName.encode('cp1257')
        
        param10 = arcpy.Parameter(
            displayName="Naudojama rezultato DB su topologija",
            name="topo_strukt",
            datatype="GPBoolean",
            parameterType="Optional",
            direction="Input")
        if is_arcmap:
            param10.displayName = param10.displayName.encode('cp1257')
        param10.value = False

        param2 = arcpy.Parameter(
            displayName="Nurodomas TEDR struktūrą atitinkantis šabloninis duomenų failas (FGDB)",
            name="TEDR_TEMPLATE",
            datatype="DEWorkspace",
            parameterType="Required",
            direction="Input")
        if is_arcmap:
            param2.displayName = "Nurodomas TEDR struktūrą atitinkantis sabloninis duomenų failas (FGDB)"
            param2.displayName = param2.displayName.encode('cp1257')

        param3 = arcpy.Parameter(
            displayName="Konfigūracijos lentelė (Excel)",
            name="CONFIG",
            datatype="DEFile",
            parameterType="Required",
            direction="Input")
        param3.filter.list = ["xlsx"]
        if is_arcmap:
            param3.displayName = param3.displayName.encode('cp1257')

        param3.value = os.path.abspath(os.path.join(current_dir, '..','konfiguracija', XLSXconfig_name))

        param4 = arcpy.Parameter(
            displayName="Rezultato atributas",
            name="rez_attrib",
            datatype="GPString",
            parameterType="Optional",
            direction="Input")
        if is_arcmap:
            param4.displayName = param4.displayName.encode('cp1257')

        param5 = arcpy.Parameter(
            displayName="Tvarkytojo, kurio duomenys perkeliami, ID",
            name="TID",
            datatype="GPLong",
            parameterType="Optional",
            direction="Input")
        if is_arcmap:
            param5.displayName = param5.displayName.encode('cp1257')

        param6 = arcpy.Parameter(
            displayName="Šulinių kortelių failai - katalogas",
            name="MANHOLE_FROM_DIR",
            datatype="DEWorkspace",
            parameterType="Optional",
            direction="Input")
        if is_arcmap:
            param6.displayName = param6.displayName.encode('cp1257')

        param7 = arcpy.Parameter(
            displayName="Gaunamas kortelių katalogas",
            name="MANHOLE_TO_DIR",
            datatype="DEWorkspace",
            parameterType="Optional",
            direction="Input")
        if is_arcmap:
            param7.displayName = param7.displayName.encode('cp1257')

        param8 = arcpy.Parameter(
            displayName="Šulinio kortelės failo plėtinys",
            name="EXTENSION",
            datatype="GPString",
            parameterType="Optional",
            direction="Input")
        param8.value = "jpg;pdf;png"
        if is_arcmap:
            param8.displayName = param8.displayName.encode('cp1257')

        param9 = arcpy.Parameter(
            displayName="Log failo vieta",
            name="log_file",
            datatype="GPString",
            parameterType="Derived",
            direction="Output")
        if is_arcmap:
            param9.displayName = param9.displayName.encode('cp1257')

        params = [param0, param1, param3, param10, param5, param6, param7, param8, param9]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        parameters[4].clearMessage()
        rez_attr_val = parameters[4].valueAsText
        if rez_attr_val:
            if rez_attr_val[0].isnumeric():
                parameters[4].setErrorMessage(
                    "Atributo pavadinimas negali prasidėti skaičiumi")
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        convert(
            parameters[0].valueAsText,
            parameters[1].valueAsText,
            parameters[2].valueAsText,
            parameters[3].value,
            rez_attrib_name,
            parameters[4].valueAsText,
            parameters[5].value,
            parameters[6].valueAsText,
            parameters[7].valueAsText
        )



def init_loggers(logger_name, log_file_location, datestr):
    # Logging initialization
    global info_logger
    global err_logger
    info_logger_name = logger_name + "_" + datestr + "_INFO.log"
    err_logger_name = logger_name + "_" + datestr + "_WARNING.log"
    if not arcpy.Exists(log_file_location):
        os.makedirs(log_file_location)
    info_logger = logging.getLogger(info_logger_name)
    err_logger = logging.getLogger(err_logger_name)

    log_file_path = ""
    err_log_file_path = ""

    if not len(info_logger.handlers):
        log_file_path = create_logger(info_logger, log_file_location, info_logger_name, logging.INFO)
    
    if not len(err_logger.handlers):
        err_log_file_path = create_logger(err_logger, log_file_location, err_logger_name, logging.WARNING)

    return log_file_path, err_log_file_path


def create_logger(logger, log_file_location, logger_name, level):
    logger.setLevel(level)
    formatter = logging.Formatter('%(asctime)s\t%(levelname)s\t%(message)s')
    log_file_path = os.path.join(log_file_location, logger_name)
    log_handler = handlers.TimedRotatingFileHandler(log_file_path, when="D", interval=1, backupCount=10, encoding="utf-8")
    log_handler.setLevel(level)
    log_handler.setFormatter(formatter)
    logger.addHandler(log_handler)
    
    # manual rollover counting from first log entry date, not from last edit
    if os.path.exists(log_handler.baseFilename):
        with open(log_handler.baseFilename) as f:
            first_line = f.readline()
    
        if first_line:
            pattern = r"([\d: -]*).*( -)"
            date_time_str = re.match(pattern, first_line).group(1)
            date_time_obj = datetime.strptime(date_time_str, logging.Formatter.default_time_format)
            if date_time_obj.timestamp() + log_handler.interval < datetime.now().timestamp():
                log_handler.doRollover()
    return log_file_path


def log(msg, layer="", code="", level = "INFO"):
    if is_arcmap:
        msg = LTtoLatin(msg)
    msg = "{}\t{}\t{}".format(layer, code, msg)
    arcpy.AddMessage(msg)

    if level == "ERROR":
        info_logger.error(msg, exc_info=True)
        err_logger.error(msg, exc_info=True)
    elif level == "WARNING":
        info_logger.warning(msg)
        err_logger.warning(msg)
    else:
        info_logger.info(msg)


def convert(SEDR, TEDR_dir, config, topo_DB, rez_attrib, tid, manhole_from_dir=None, manhole_to_dir=None, extensions="jpg"):
    # Panaikinama: TEDR_template, config, rez_attrib
    global gedr_field_types
    global edit_objects

    gedr_field_types = {}
    edit_objects = {}

    rez_attrib = "perkelta"
    TEDR_template = os.path.join(current_dir, '..','konfiguracija', wsXML_topo)
    if topo_DB == False:
        TEDR_template = os.path.join(current_dir, '..','konfiguracija',wsXML_betopo)

    # config = os.path.join(current_dir, '..','konfiguracija', XLSXconfig_name)

    outputs_location = TEDR_dir
    datestr = datetime.now().strftime("%Y-%m-%d_%H%M%S")
    log_file_path, err_log_file_path = init_loggers("SEDRtoTEDR", outputs_location, datestr)
    log("Įrankio darbo pradžia.")
    log("Duoti parametrai:")
    log("SEDR failas: {}".format(SEDR))
    log("Gaunamo failo direktorija: {}".format(TEDR_dir))
    log("TEDR struktūros šablonas: {}".format(TEDR_template))
    log("Konfigūracijos lentelė: {}".format(config))
    log("Rezultato atributas: {}".format(rez_attrib))
    log("Tvarkytojo ID: {}".format(tid))

    if tid:
        tid = int(tid)

    config_props = read_xls(config, "fcmap")
    if config_props is None:
        return
    
    card_extensions = extensions.upper().split(";")
    
    # TODO: uncomment
    TEDR_file = copy_template_to_tedr(TEDR_template, TEDR_dir, datestr)
    
    # Prieš pridedant laukus, reikia išjungti redagavimo sekimą, nes sugadinami last updated date laukai
    disable_editor_tracking(SEDR)


    # TODO: uncomment
    if not add_rez_field(SEDR, rez_attrib):
        log("Įrankis baigė darbą.")
        return

    # remove source spatial index to avoid unexpected errors
    # not adding new ones, because it creates another error sometimes
    remove_spatial_indexes(SEDR)
    
    try:
        res_moved = {}
        res_failed = []
        sum_moved = 0
        
        for item in config_props:
            moved, failed = move_features(item, SEDR, TEDR_file, rez_attrib, tid)
            if len(failed) > 0:
                res_failed += failed
            to_fc = item["ToFC"]
            if to_fc in res_moved:
                res_moved[to_fc] += moved
            else:
                res_moved[to_fc] = moved
            sum_moved += moved
        
        for edit_obj in edit_objects.values():
            edit_obj.stopEditing(True)

        if manhole_from_dir and manhole_to_dir:
            cards_moved, cards_failed = move_manhole_cards(manhole_from_dir, manhole_to_dir, card_extensions)

        # to json
        moved_path = os.path.join(outputs_location, "moved_{}.json".format(datestr))
        failed_path = os.path.join(outputs_location, "failed_{}.json".format(datestr))
        with open(moved_path, "w") as outfile:
            json.dump(res_moved, outfile)
        with open(failed_path, "w") as outfile:
            json.dump(res_failed, outfile)
        arcpy.SetParameterAsText(8, log_file_path)
        if is_arcmap:
            log("VISO perkelta objektu: {}".format(sum_moved))
            log("Neperkelta del klaidu: {}".format(len(res_failed)))
        else:
            log("VISO perkelta objektų: {}".format(sum_moved))
            log("Neperkelta dėl klaidų: {}".format(len(res_failed)))            
        log_failed_by_layer(SEDR, rez_attrib)
        if manhole_from_dir and manhole_to_dir:
            log("VISO perkelta kortelių: {}".format(cards_moved))
            log("Dėl klaidų neperkeltų kortelių: {}".format(cards_failed))

        log("Log failai: {}; {}".format(log_file_path, err_log_file_path))
        log("Sukurta TEDR FGDB: {}".format(TEDR_file))
        # log("Įrankis baigė darbą.")
    except Exception as e:
        log("Netikėta klaida perkeliant", level="ERROR")
        arcpy.AddError(e)
    
    log("Įrankis baigė darbą.")


def add_rez_field(SEDR, rez_attrib):
    try:
        add_field_to_sedr(SEDR, rez_attrib)
    except Exception as e:
        log("Klaida pridedant papildomą atributą '{}'".format(rez_attrib), level="ERROR")
        arcpy.AddError(e)
        return False

    return True


def log_failed_by_layer(gdb, rez_attrib):
    arcpy.env.workspace = gdb
    datasets = arcpy.ListDatasets(feature_type='feature')
    datasets = [''] + datasets if datasets is not None else []
    fails = {}
    for ds in datasets:
        for fc in arcpy.ListFeatureClasses(feature_dataset=ds):
            fc_path = os.path.join(gdb, ds, fc)
            if "KODAS" in [f.name.upper() for f in arcpy.ListFields(fc_path)]:
                fields = ["KODAS"]
                has_code = True
            else:
                fields = ["OBJECTID"]
                has_code = False
            try:
                with arcpy.da.SearchCursor(fc_path, fields, "{} IS NULL".format(rez_attrib)) as cur:
                    for row in cur:
                        if has_code:
                            code = row[0]
                        else:
                            code = ""

                        if fc not in fails:
                            fails[fc] = {}

                        if code not in fails[fc]:
                            fails[fc][code] = 0

                        fails[fc][code] += 1
            except Exception as e:
                arcpy.AddMessage(e)
    if len(fails) == 0:
        log("Bendrai neperkelta pradinių objektų: 0")
        return

    log("Bendrai neperkelta pradinių objektų:")
    for fail_fc in fails:
        for code in fails[fail_fc]:
            count = fails[fail_fc][code]
            log("Dataset {} \t {} \t {}".format(fail_fc, code, count))


def move_manhole_cards(manhole_from_dir, manhole_to_dir, card_extensions):
    log("Kopijuojamos šulinių kortelės")
    moved = 0
    failed = 0
    if not arcpy.Exists(manhole_from_dir):
        log("Nurodyta šulinių kortelių paieškos direktorija neegzistuoja")
        return moved, failed
    if not arcpy.Exists(manhole_to_dir):
        log("Gaunama šulinių kortelių direktorija neegzistuoja")
        return moved, failed
    
    for root, dirs, files in os.walk(manhole_from_dir):
        target_dir = root.replace(manhole_from_dir, manhole_to_dir)
        for name in files:
            if name.split(".")[-1].upper() in card_extensions:
                src_path = os.path.join(root, name)
                target_path = os.path.join(target_dir, name)
                if not os.path.exists(target_dir):
                    os.mkdir(target_dir)
                if os.path.exists(target_path):
                    log("Kortelė {} jau egzistuoja.".format(target_path))
                    failed += 1
                else:
                    shutil.copy(src_path, target_path)
                    log("Perkelta kortelė: {}".format(target_path))
                    moved += 1
    log("Iš viso perkelta kortelių: {}".format(moved))
    log("Iš viso neperkelta kortelių: {}".format(failed))
    return moved, failed
                

def copy_template_to_tedr(TEDR_template, TEDR_dir, datestr):  
    log("Kuriama DB pagal šabloną")
    TEDR_name = "SEDR_TEDR_{}.gdb".format(datestr)
    TEDR_path = os.path.join(TEDR_dir, TEDR_name)

    # arcpy.Copy_management(TEDR_template, TEDR_path)

    arcpy.management.CreateFileGDB(TEDR_dir, TEDR_name)
    arcpy.management.ImportXMLWorkspaceDocument(TEDR_path, TEDR_template)


    return TEDR_path

def disable_editor_tracking(sedr):
	
	arcpy.env.workspace = sedr

	# Get all the stand alone tables and feature classes
	standaloneList = arcpy.ListTables() + arcpy.ListFeatureClasses()

	# Execute enable editor tracking on standalone tables
	for dataset in standaloneList:
		log('Išjungiamas redagavimo sekimas: {}'.format(dataset))
		arcpy.management.DisableEditorTracking(dataset)

	# For feature datasets get all of the featureclasses and add them to the list
	for dataset in arcpy.ListDatasets("", "Feature"):
		arcpy.env.workspace = os.path.join(sedr,dataset)
		for fc in arcpy.ListFeatureClasses():
			log('Išjungiamas redagavimo sekimas: {}'.format(fc))
			arcpy.management.DisableEditorTracking(fc)


def add_field_to_sedr(sedr, new_attrib):
    # log("TEDR rezultatui pridedamas papildomas atributas '{}'".format(rez_attrib))
    log("SEDR rinkiniui pridedamas papildomas atributas '{}'".format(new_attrib))
    arcpy.env.workspace = sedr
    datasets = arcpy.ListDatasets(feature_type='feature')
    datasets = [''] + datasets if datasets is not None else []
    for ds in datasets:
        for fc in arcpy.ListFeatureClasses(feature_dataset=ds):
            fc_path = os.path.join(sedr, ds, fc)
            if new_attrib in [f.name for f in arcpy.ListFields(fc_path)]:
                log("Sluoksnis jau turi lauką '{}'. Trinama.".format(new_attrib), fc, "", "WARNING")
                arcpy.DeleteField_management(fc_path, new_attrib)
            # else:
            arcpy.AddField_management(fc_path, new_attrib, "TEXT")
            log("Sluoksnyje sukurtas laukas '{}'.".format(new_attrib), fc)


def remove_spatial_indexes(sedr):
    log("SEDR rinkiniui pašalinami erdviniai indeksai")
    start_time = datetime.now()
    arcpy.env.workspace = sedr
    datasets = arcpy.ListDatasets(feature_type='feature')
    datasets = [''] + datasets if datasets is not None else []
    for ds in datasets:
        for fc in arcpy.ListFeatureClasses(feature_dataset=ds):
            fc_path = os.path.join(sedr, ds, fc)
            try:
                arcpy.RemoveSpatialIndex_management(fc_path)
            except Exception as e:
                pass
    log("Perskaičiavimo trukmė: {}".format(datetime.now() - start_time))


def read_xls(xlsfilePath, sheet_name):
    log("Nuskaitoma konfigūracija")
    xlsbook = xlrd.open_workbook(xlsfilePath)
    propList = []
    wsheet = xlsbook.sheet_by_name(sheet_name)
    unique_names = []

    # nuskaityti laukų pavadinimus ir sudėlioti seką kaip išdėlioti nuskaitymo tvarką
    colIndexDict = {}
    for col in range(wsheet.ncols):
        colname = wsheet.cell_value(0,col)
        colIndexDict[colname.lower()]= col
    # arcpy.AddMessage("Cadprops COLUMNS: {}".format(colIndexDict))

    if sheet_name == "fcmap":
        for rn in range(1,wsheet.nrows):
            propsDict = {}
            propsDict['FromFC'] = wsheet.cell_value(rn,colIndexDict.get('fromfc'))
            propsDict['ToFC'] = wsheet.cell_value(rn,colIndexDict.get('tofc'))
            try:
                propsDict['FromKodas'] = int(wsheet.cell_value(rn,colIndexDict.get('fromkodas')))
            except:
                propsDict['FromKodas'] = ""
            propsDict['ToKodas'] = int(wsheet.cell_value(rn,colIndexDict.get('tokodas')))
            json_mp = wsheet.cell_value(rn,colIndexDict.get('attrmapping'))
            if json_mp:
                try:
                    propsDict['AttrMapping'] = json.loads(json_mp)
                except Exception as e:
                    log("Klaida konfigūracijos faile skaitant stulpelį 'AttrMapping'. Eilutė: {}.".format(rn+1), level="ERROR")
                    arcpy.AddError(e)
                    return None
            else:
                propsDict['AttrMapping'] = None
            propsDict['FileMapping'] = wsheet.cell_value(rn,colIndexDict.get('filemapping'))
            
            # praleidžiamos pasikartojančios taisyklės
            unique_name = "{}, {}, {}, {}".format(propsDict['FromFC'].lower(), propsDict['ToFC'].lower(),
                propsDict['FromKodas'], propsDict['ToKodas'])
            if unique_name in unique_names:
                log("Praleidžiama pasikartojanti konfigūracijos eilutė: {}".format(unique_name), level="WARNING")
            else:
                unique_names.append(unique_name)

            # cadPropsDict['pavad'] = wsheet.cell_value(rn,colIndexDict.get('pavad')) if colIndexDict.get('pavad') else None
            propList.append(propsDict)
    elif sheet_name == "fieldmap":
        pass

    return propList


def move_features(item, SEDR, TEDR_file, rez_attrib, tid):
    global edit_objects

    s_fc = item["FromFC"]
    t_fc = item["ToFC"]
    if item["FromKodas"]:
        s_where = "Kodas = {}".format(item["FromKodas"])
    else:
        s_where = None

    # if item["FromKodas"] == "3801":
    #     pass
    # log("Perkeliamas pradinis kodas {}.  iš {} į {}:".format(s_fc, t_fc))
    src_path = os.path.join(SEDR, s_fc)
    trg_path = os.path.join(TEDR_file, t_fc)

    s_fc_name = os.path.basename(s_fc)
    moved = 0
    failed = []

    if not item["FromFC"] or not arcpy.Exists(src_path):
        log("Nerastas SEDR sluoksnis. Perkėlimas praleidžiamas.", s_fc_name, item["FromKodas"], "WARNING")
        return moved, failed

    if not item["ToKodas"]:
        error = "Sluoksnio perkėlimo taisyklėje nenurodytas 'toKodas'. Perkėlimas negalimas."
        log(error, s_fc_name, item["FromKodas"], "WARNING")
        failed.append([s_fc, 0, error])
        return moved, failed

    possible_source_fields = []
    objid_fld_name = ""
    for f in arcpy.ListFields(src_path):
        if f.type == "OID":
            objid_fld_name = f.name.upper()
        possible_source_fields.append(f.name.upper())
    wrong_source_fields = []

    attr_config = item["AttrMapping"]
    from_fields = []
    to_fields = []
    to_val_config = {}
    bad_config = False
    for f in attr_config:
        if f[0]:
            if f[0].upper() not in possible_source_fields:
                wrong_source_fields.append(f[0])
                bad_config = True
                continue

            if f[0].lower() in ["kodas"]:
                continue

            if f[1]:
                from_fields.append(f[0])
                to_fields.append(f[1])
                temp = {}
                for vals in f[2]:
                    temp[vals[0]] = vals[1]
                to_val_config[f[1]] = temp
            else:
                log("Atributui '{}' nenurodytas naujas pavadinimas.".format(f[0]), s_fc_name, item["FromKodas"], "WARNING")
                # failed.append([s_fc, 0, "Kodas: {}. Atributui '{}' nenurodytas naujas pavadinimas.".format(item["FromKodas"], f[0])])
                bad_config = True
    
    if not item["FromKodas"]:
        additional_fields = ["SHAPE@"]
    else:
        additional_fields = ["SHAPE@", "Kodas"]

    from_fields = from_fields + additional_fields + [objid_fld_name, rez_attrib]

    if not item["FromKodas"]:
        additional_to_fields = ["Kodas", "TID", "UID", "Sinch"]
    else:
        additional_to_fields = ["TID", "UID", "Sinch"]
    # to_fields = to_fields + additional_fields + additional_to_fields
    added_flds_count = 0
    to_fields = to_fields + additional_fields
    for atf in additional_to_fields:
        if atf not in to_fields:
            to_fields.append(atf)
            added_flds_count += 1

    possible_target_fields = [f.name.upper() for f in arcpy.ListFields(trg_path)]
    ok_to_fields = []
    wrong_fields = []
    ids_to_remove = []
    for tf in to_fields:
        if tf.upper() not in possible_target_fields and tf.upper() != "SHAPE@":
            wrong_fields.append(tf)
            wrong_idx = to_fields.index(tf)
            ids_to_remove.append(wrong_idx)
        else:
            ok_to_fields.append(tf)
    
    ok_from_fields = []
    for idx, fld_name in enumerate(from_fields):
        if idx not in ids_to_remove:
            ok_from_fields.append(fld_name)

    
    if len(wrong_source_fields) > 0:
        error = "Šaltinio duomenyse nėra atributų {}. Sluoksnis: '{}'.".format(wrong_source_fields, s_fc)
        log(error, s_fc_name, item["FromKodas"], "WARNING")
    
    if len(wrong_fields) > 0:
        error = "Gaunamuose duomenyse nėra atributų {}. Sluoksnis: '{}'.".format(wrong_fields, t_fc)
        log(error, s_fc_name, item["FromKodas"], "WARNING")
        # failed.append([s_fc, 0, error])
        # return moved, failed

    # Ar stabdyti jei config su klaidomis?
    # if bad_config:
    #     return moved, failed

    # add shape ir kodas later
    
    if TEDR_file in edit_objects:
        edit_tedr = edit_objects[TEDR_file]
    else:
        edit_tedr = arcpy.da.Editor(TEDR_file)
        edit_tedr.startEditing(False, False)
        edit_objects[TEDR_file] = edit_tedr
    edit_tedr.startOperation()

    obj_id_idx = ok_from_fields.index(objid_fld_name)
    rez_attrib_idx = ok_from_fields.index(rez_attrib)
    imported_obj_ids = []

    with arcpy.da.SearchCursor(src_path, ok_from_fields, s_where) as s_cur:
        with arcpy.da.InsertCursor(trg_path, ok_to_fields) as t_cur:
            try:
                for s_row in s_cur:
                    s_row = list(s_row)
                    # cut out values that will go to target
                    common_len = len(ok_to_fields) - added_flds_count
                    new_row = s_row[:common_len]
                    for afc in range(added_flds_count):
                        new_row.append(None)

                    obj_id = s_row[obj_id_idx]
                    
                    new_row = set_additional_values(new_row, to_val_config, ok_to_fields, tid, item["ToKodas"])

                    from_field_types = get_gedr_field_types(SEDR, s_fc)
                    to_field_types = get_gedr_field_types(TEDR_file, t_fc)
                    new_row = check_types_and_values(new_row, ok_from_fields, ok_to_fields, from_field_types,
                        to_field_types, s_fc, obj_id)
                    
                    try:
                        insert_to_target(t_cur, new_row, ok_to_fields, s_fc, obj_id, item["FromKodas"], tid)
                        moved += 1
                        s_row[rez_attrib_idx] = t_fc
                        imported_obj_ids.append(obj_id)
                    except Exception as e:
                        failed.append([s_fc, obj_id, str(e), item["FromKodas"]])

            except Exception as e:
                if is_arcmap: log("Klaida perkeliant irasus", s_fc_name, item["FromKodas"], "ERROR")
                else: log("Klaida perkeliant įrašus", s_fc_name, item["FromKodas"], "ERROR")
    
    edit_tedr.stopOperation()

    # update source to mark features that were successfully moved
    if SEDR in edit_objects:
        edit = edit_objects[SEDR]
        if not edit.isEditing:
            edit.startEditing(False, False)
    else:
        edit = arcpy.da.Editor(SEDR)
        edit.startEditing(False, False)
        edit_objects[SEDR] = edit
    edit.startOperation()

    with arcpy.da.UpdateCursor(src_path, [objid_fld_name, rez_attrib], s_where) as s_cur:
        for s_row in s_cur:
            if s_row[0] not in imported_obj_ids:
                continue
            s_row[1] = t_fc
            s_cur.updateRow(s_row)
    edit.stopOperation()

    # wait to be sure lock files are deleted
    time.sleep(2)

    if is_arcmap: log("Perkelta {} irasu.".format(moved), s_fc_name, item["FromKodas"])
    else: log("Perkelta {} įrašų.".format(moved), s_fc_name, item["FromKodas"])
    # log(datetime.now() - startas)
    return moved, failed


def insert_to_target(cur, row, fields, s_fc, obj_id, code, tid):
    shp_idx = fields.index("SHAPE@")
    if row[shp_idx].partCount > 1:
        uid_idx = fields.index("UID")
        split_geoms = split_multipart_geometry(row[shp_idx])
        for geom in split_geoms:
            row[shp_idx] = geom
            row[uid_idx] = generate_uid(tid)
            cur.insertRow(row)
        log("Daugiadalis objektas išskaidytas {}:{}".format(s_fc, obj_id), s_fc, code, "WARNING" )
    else:
        # if row[shp_idx].area == 0:
        #     print("Geometry area is 0 (single part)")
        cur.insertRow(row)


def split_multipart_geometry(geometry):
    coords = json.loads(geometry.JSON)
    split = []
    if geometry.type == "polygon":
        # print(coords)
        # 'rings' has no distinction between interior and exterior rings
        if "curveRings" in coords:
            rings = coords["curveRings"]
        else:
            rings = coords["rings"]
        
        ring_idx = 0
        partno = 0
        for part in geometry:
            partno += 1
            rings_in_part = [rings[ring_idx]]
            for pnt in part:
                if not pnt:
                    # If pnt is None, this represents an interior ring
                    # take one from 'rings' list
                    ring_idx += 1
                    rings_in_part.append(rings[ring_idx])
            ring_idx += 1

            # using 'curveRings', because it changes to 'rings' if no curves are in the list
            esri_json = {
                "curveRings": rings_in_part,
                "spatialReference": coords["spatialReference"]
            }
            
            new_shape = arcpy.AsShape(esri_json, True)
            split.append(new_shape)
            if new_shape.area == 0:
                print("Geometry area is 0 (multipart)")
                print(coords)
    elif geometry.type == "point":
        for pnt in geometry:
            split.append([pnt.X, pnt.Y])
    else:
        if "curvePaths" in coords:
            paths = coords["curvePaths"]
        else:
            paths = coords["paths"]
        for path in paths:
            esri_json = {
                "curvePaths":[path],
                "spatialReference": coords["spatialReference"]
            }
            new_shape = arcpy.AsShape(esri_json, True)
            split.append(new_shape)
    return split


def set_additional_values(new_row, to_val_config, ok_to_fields, tid, code):
    for idx, fld_name in enumerate(ok_to_fields):
        if fld_name == "SHAPE@":
            continue
        if fld_name.lower() == "tid":
            new_row[idx] = tid
            continue
        if fld_name.lower() == "sinch":
            new_row[idx] = 1
            continue
        if fld_name.lower() == "kodas":
            new_row[idx] = code
            continue
        if fld_name.lower() == "uid":
            if not new_row[idx]:
                new_row[idx] = generate_uid(tid)
            continue
        val_cfg = to_val_config[fld_name]
        cur_value = str(new_row[idx]) if new_row[idx] is not None else None
        if cur_value in val_cfg:
            new_row[idx] = val_cfg[cur_value]
    return new_row


def generate_uid(tid):
    global last_generated_uid

    uid = str(uuid.uuid1())
    while last_generated_uid == uid:
        uid = str(uuid.uuid1())
    last_generated_uid = uid

    if tid and tid < 4096:
        hexed = str(hex(tid))[2:]
        if len(hexed) < 3:
            hexed = hexed.zfill(3)
        result = uid[:5] + hexed + uid[8:]
        return "{" + result + "}"
    else:
        return "{" + uid + "}"


def get_gedr_field_types(gdb, ds_fc):
    global gedr_field_types
    try:
        prefix = os.path.basename(gdb)
        return gedr_field_types[prefix + "_" + ds_fc]
    except KeyError:
        fc_path = os.path.join(gdb, ds_fc)
        fc_types = {}
        for f in arcpy.ListFields(fc_path):
            if f.name.upper():
                fc_types[f.name.upper()] = [f.type, f.length]
        gedr_field_types[prefix + "_" + ds_fc] = fc_types
        return fc_types


def check_types_and_values(row, source_fields, target_fields, from_field_types, to_field_types, fc_name, obj_id):
    for index, rw in enumerate(row):
        if row[index] is None:
            continue
        t_fld_name = target_fields[index].upper()
        # Praleidžiami atskirai apdoroti laukai
        if t_fld_name in ["SHAPE@", "KODAS", "TID"]:
            continue

        try:
            s_fld_name = source_fields[index].upper()
        except IndexError as e:
            # ok if one of these fields
            if t_fld_name in ["UID", "SINCH"]:
                continue
            else:
                raise(e)
        
        if s_fld_name in from_field_types and t_fld_name in to_field_types:
            type_from = from_field_types[s_fld_name][0]
            type_to = to_field_types[t_fld_name][0]
            if type_from != type_to:
                try:
                    row[index] = convert_value(row[index], type_to)
                except:
                    row[index] = None
                    log("Nesuderinamas lauko {} tipas pakeistas NULL reikšme.".format(s_fld_name), fc_name, obj_id, "WARNING")
            
            if type_to == "String" and row[index]:
                # Apkerpami per ilgi tekstai
                max_len = to_field_types[t_fld_name][1]
                if len(row[index]) > max_len:
                    row[index] = row[index][:max_len]
                    log("Atributinio lauko reikšmė {} sutrumpinta pagal lauko ilgį.".format(s_fld_name), fc_name, obj_id, "WARNING")
        else:
            log("Lyginant laukų tipus nerastas vienas iš laukų: {} {}".format(s_fld_name, t_fld_name))
    return row


def convert_value(value, type_to):
    if value is None:
        return None
    
    if type_to == "String":
        return str(value)
    elif type_to == "SmallInteger":
        if type(value) == "float":
            return None
        return int(value)
    elif type_to == "Integer":
        if type(value) == "float":
            return None
        return long(value)
    elif type_to in ["Single", "Double"]:
        return float(value)
    else:
        return value


def LTtoLatin(myInput):
    if sys.version_info[0] < 3:
        # Standartiniai koduočių metodai neveikia naudojant arcmap
        chars = ["ą",      "č",     "ę",       "ė",      "į",      "š",      "ų",      "ū",      "ž",
                "Ą",       "Č",     "Ę",       "Ė",      "Į",      "Š",      "Ų",      "Ū",      "Ž",
                "ÄÆ","s³","s«","s¾"]
        lat = ["a", "c", "e", "e", "į", "s", "u", "u", "z",
                "A",  "C", "E", "E", "I", "S", "U", "U", "Z",
                'i','u','u',"z"]

        for index, char in enumerate(chars):
            myInput = myInput.replace(char, lat[index])
    return myInput